--------------------------------------------------------------------------------
-- Company: 
-- Engineer:
--
-- Create Date:    02:19:35 11/09/05
-- Design Name:    
-- Module Name:    stream - Behavioral
-- Project Name:   
-- Target Device:  
-- Tool versions:  
-- Description:
--
-- Dependencies:
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_SIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity stream is
  port (
    CLK : in std_logic;
    funct : in  std_logic_vector(3 downto 0);
    put_left_word, put_right_word : in  std_logic_vector(31 downto 0);
    shift : in  std_logic_vector(4 downto 0);
    get_comp_word, get_simp_word : out  std_logic_vector(31 downto 0);
    complete : out std_logic;
    USBD : inout std_logic_vector(7 downto 0);
    USBRDX : out std_logic;
    USBWR : out std_logic;
    USBTXEX : in std_logic;
    USBRXFX : in std_logic);
end stream;

architecture Behavioral of stream is
  signal integer_buf : std_logic_vector(31 downto 0);
  signal sign_buf : std_logic;
  signal exponent_buf : std_logic_vector(8 downto 0);
  signal fraction_buf : std_logic_vector(47 downto 0);
  signal complete_buf : std_logic;
  signal count : std_logic_vector(2 downto 0) := "000";

  procedure split (
    constant word     : in  std_logic_vector(31 downto 0);
    variable sign     : out std_logic;
    variable exponent : out std_logic_vector(7 downto 0);
    variable fraction : out std_logic_vector(23 downto 0)) is
  begin  -- split
    sign := word(31);
    exponent(7 downto 1) := word(30 downto 24);
    exponent(0) := not (word(30) or word(29) or word(28) or word(27) or word(26) or word(25) or word(24)) or word(23);
    fraction(23) := word(30) or word(29) or word(28) or word(27) or word(26) or word(25) or word(24) or word(23);
    fraction(22 downto 0) := word(22 downto 0);
  end split;
  
  procedure join (
    signal word       : out std_logic_vector(31 downto 0);
    constant sign     : in  std_logic;
    constant exponent : in  std_logic_vector(8 downto 0);
    constant fraction : in  std_logic_vector(47 downto 0)) is
    variable shift : std_logic_vector(4 downto 0);
    variable shift_sign : std_logic;
    variable shift_exponent : std_logic_vector(8 downto 0);
    variable shift_fraction : std_logic_vector(46 downto 0);    
  begin  -- join
    
--    floor support
-- 
--    shift(4) := not fraction(47) and not fraction(46) and not fraction(45) and not fraction(44) and not fraction(43) and not fraction(42) and not fraction(41) and not fraction(40) and not fraction(39) and not fraction(38) and not fraction(37) and not fraction(36) and not fraction(35) and not fraction(34) and not fraction(33) and not fraction(32);
--    shift(3) := not fraction(47) and not fraction(46) and not fraction(45) and not fraction(44) and not fraction(43) and not fraction(42) and not fraction(41) and not fraction(40) and (fraction(39) or fraction(38) or fraction(37) or fraction(36) or fraction(35) or fraction(34) or fraction(33) or fraction(32) or (not fraction(31) and not fraction(30) and not fraction(29) and not fraction(28) and not fraction(27) and not fraction(26) and not fraction(25) and not fraction(24)));
--    shift(2) := not fraction(47) and not fraction(46) and not fraction(45) and not fraction(44) and (fraction(43) or fraction(42) or fraction(41) or fraction(40) or (not fraction(39) and not fraction(38) and not fraction(37) and not fraction(36) and (fraction(35) or fraction(34) or fraction(33) or fraction(32) or (not fraction(31) and not fraction(30) and not fraction(29) and not fraction(28) and (fraction(27) or fraction(26) or fraction(25) or fraction(24) or (not fraction(23) and not fraction(22) and not fraction(21) and not fraction(20)))))));
--    shift(1) := not fraction(47) and not fraction(46) and (fraction(45) or fraction(44) or (not fraction(43) and not fraction(42) and (fraction(41) or fraction(40) or (not fraction(39) and not fraction(38) and (fraction(37) or fraction(36) or (not fraction(35) and not fraction(34) and (fraction(33) or fraction(32) or (not fraction(31) and not fraction(30) and (fraction(29) or fraction(28) or (not fraction(27) and not fraction(26) and (fraction(25) or fraction(24) or (not fraction(23) and not fraction(22) and (fraction(21) or fraction(20) or (not fraction(19) and not fraction(18)))))))))))))));
--    shift(0) := not fraction(47) and (fraction(46) or (not fraction(45) and (fraction(44) or (not fraction(43) and (fraction(42) or (not fraction(41) and (fraction(40) or (not fraction(39) and (fraction(38) or (not fraction(37) and (fraction(36) or (not fraction(35) and (fraction(34) or (not fraction(33) and (fraction(32) or (not fraction(31) and (fraction(30) or (not fraction(29) and (fraction(28) or (not fraction(27) and (fraction(26) or (not fraction(25) and (fraction(24) or (not fraction(23) and (fraction(22) or (not fraction(21) and (fraction(20) or(not fraction(19) and (fraction(18) or not fraction(17))))))))))))))))))))))))))))));

    shift(4) := not fraction(47) and not fraction(46) and not fraction(45) and not fraction(44) and not fraction(43) and not fraction(42) and not fraction(41) and not fraction(40) and not fraction(39) and not fraction(38) and not fraction(37) and not fraction(36) and not fraction(35) and not fraction(34) and not fraction(33) and not fraction(32);
    shift(3) := not fraction(47) and not fraction(46) and not fraction(45) and not fraction(44) and not fraction(43) and not fraction(42) and not fraction(41) and not fraction(40) and (fraction(39) or fraction(38) or fraction(37) or fraction(36) or fraction(35) or fraction(34) or fraction(33) or fraction(32) or (not fraction(31) and not fraction(30) and not fraction(29) and not fraction(28) and not fraction(27) and not fraction(26) and not fraction(25) and not fraction(24)));
    shift(2) := not fraction(47) and not fraction(46) and not fraction(45) and not fraction(44) and (fraction(43) or fraction(42) or fraction(41) or fraction(40) or (not fraction(39) and not fraction(38) and not fraction(37) and not fraction(36) and (fraction(35) or fraction(34) or fraction(33) or fraction(32) or (not fraction(31) and not fraction(30) and not fraction(29) and not fraction(28) and (fraction(27) or fraction(26) or fraction(25) or fraction(24))))));
    shift(1) := not fraction(47) and not fraction(46) and (fraction(45) or fraction(44) or (not fraction(43) and not fraction(42) and (fraction(41) or fraction(40) or (not fraction(39) and not fraction(38) and (fraction(37) or fraction(36) or (not fraction(35) and not fraction(34) and (fraction(33) or fraction(32) or (not fraction(31) and not fraction(30) and (fraction(29) or fraction(28) or (not fraction(27) and not fraction(26) and (fraction(25) or fraction(24))))))))))));
    shift(0) := not fraction(47) and (fraction(46) or (not fraction(45) and (fraction(44) or (not fraction(43) and (fraction(42) or (not fraction(41) and (fraction(40) or (not fraction(39) and (fraction(38) or (not fraction(37) and (fraction(36) or (not fraction(35) and (fraction(34) or (not fraction(33) and (fraction(32) or (not fraction(31) and (fraction(30) or (not fraction(29) and (fraction(28) or (not fraction(27) and (fraction(26) or (not fraction(25) and fraction(24)))))))))))))))))))))));

-- epsiron support  
-- 
--    word(31) <= sign;
--    shift_exponent := exponent - ('0' & (shift(4) and shift(3) and shift(2) and shift(1) and shift(0) and not exponent(8)) & "00" & shift);
--    if shift_exponent(8) = '0' then
--      word(30 downto 23) <= shift_exponent(7 downto 0) + 1;
--      shift_fraction := std_logic_vector(shl(unsigned(fraction(46 downto 0)), unsigned(shift)));
--    elsif shift_exponent(8) = '1' then
--      word(30 downto 23) <= "00000000"; 
--      if exponent(8) = '0' then
--        shift_fraction := std_logic_vector(shl(unsigned(fraction(46 downto 0)), unsigned(exponent(4 downto 0))));
--      elsif exponent(8) = '1' then
--        if (exponent(7) and exponent(6) and exponent(5)) = '0' then
--          shift_fraction := "X00000000000000000000000XXXXXXXXXXXXXXXXXXXXXXX";
--        elsif (exponent(7) and exponent(6) and exponent(5)) = '1' then
--          shift_fraction := std_logic_vector(shr(unsigned(fraction(47 downto 1)), unsigned(not exponent(4 downto 0))));
--        else
--          shift_fraction := "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
--        end if;
--      else
--        shift_fraction := "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
--      end if;
--      shift_fraction := "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
--    else
--      word(30 downto 23) <= "XXXXXXXX";
--      shift_fraction := "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
--    end if;

    shift_sign := sign;
    shift_exponent := exponent - ("0" & (shift(4) and shift(3) and not exponent(8)) & "00" & shift) + 1;
    if shift_exponent(8) = '0' then
      shift_exponent(7 downto 0) := shift_exponent(7 downto 0);
    elsif shift_exponent(8) = '1' then
      shift_exponent(7 downto 0) := "00000000";
    else
      shift_exponent(7 downto 0) := "XXXXXXXX";
    end if;
    shift_fraction := std_logic_vector(shl(unsigned(fraction(46 downto 0)), unsigned(shift)));
    word <= shift_sign & shift_exponent(7 downto 0) & shift_fraction(46 downto 24);

-- floor support
-- 
--    if shift_fraction(23) = '0' then
--      word <= shift_sign & shift_exponent(7 downto 0) & shift_fraction(46 downto 24);
--    elsif shift_fraction(23) = '1' then
--      word <= shift_sign & ((shift_exponent(7 downto 0) & shift_fraction(46 downto 24)) + 1);
--    else
--      word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
--    end if;
    
  end join;

begin
    
  complete_buf <= not funct(3) or (count(2) and count(1) and count(0));

  complete <= complete_buf;

  USBD <=
    "ZZZZZZZZ" when funct = "1110" else
    put_left_word(7 downto 0) when funct = "1111" else
    "XXXXXXXX";

  process (funct, put_left_word, put_right_word, integer_buf, shift, sign_buf, exponent_buf, fraction_buf, USBD)
  begin  -- process
    case funct is
      when "0000" =>
        get_comp_word <= integer_buf;
      when "0001" =>
        get_comp_word <= integer_buf;
      when "0010" =>
        get_comp_word <= shl(put_left_word, shift);
      when "0011" =>
        get_comp_word <= shr(put_left_word, shift);
      when "0100" =>
        get_comp_word <= put_left_word and put_right_word;
      when "0101" =>
        get_comp_word <= put_left_word or put_right_word;
      when "0110" =>
        get_comp_word <= put_left_word xor put_right_word;
      when "0111" =>
        get_comp_word <= put_left_word xnor put_right_word;
      when "1000" =>
        join(get_comp_word, sign_buf, exponent_buf, fraction_buf);
      when "1001" =>
        join(get_comp_word, sign_buf, exponent_buf, fraction_buf);
      when "1010" =>
        join(get_comp_word, sign_buf, exponent_buf, fraction_buf);
      when "1110" =>
        get_comp_word <= "000000000000000000000000" & USBD;
      when "1111" =>
        get_comp_word <= put_left_word;
      when others =>
        get_comp_word <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
    end case;
  end process;
  
  get_simp_word <= integer_buf;
  
  process (funct, put_left_word, put_right_word)
  begin  -- process
    if funct(0) = '0' then
      integer_buf <= put_left_word + put_right_word;
    elsif funct(0) = '1' then
      integer_buf <= put_left_word - put_right_word;
    else
      integer_buf <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
    end if;
  end process;
    
  process (CLK)
    variable left_sign, right_sign : std_logic;
    variable left_exponent, right_exponent : std_logic_vector(7 downto 0);
    variable left_fraction, right_fraction : std_logic_vector(23 downto 0);
    variable left_shift, right_shift : std_logic_vector(8 downto 0);
    variable shift_sign : std_logic;
    variable shift_exponent : std_logic_vector(7 downto 0);
    variable shift_fraction : std_logic_vector(24 downto 0);
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      split(put_left_word, left_sign, left_exponent, left_fraction);
      split(put_right_word, right_sign, right_exponent, right_fraction);
      right_sign := right_sign xor funct(0);
      left_shift := ("0" & left_exponent) - ("0" & right_exponent);
      right_shift := ("0" & right_exponent) - ("0" & left_exponent);
      if unsigned(right_shift(8) & left_fraction) >= unsigned(left_shift(8) & right_fraction) then
        shift_sign := left_sign;
        shift_exponent := left_exponent;
        if (left_shift(7) or left_shift(6) or left_shift(5)) = '0' then
          shift_fraction := "0" & std_logic_vector(shr(unsigned(right_fraction), unsigned(left_shift(4 downto 0))));
          if (left_sign xor right_sign) = '0' then            
            shift_fraction :=  ("0" & left_fraction) + shift_fraction;
          elsif (left_sign xor right_sign) = '1' then      
            shift_fraction :=  ("0" & left_fraction) - shift_fraction;
          else
            shift_fraction := "XXXXXXXXXXXXXXXXXXXXXXXXX";
          end if;
        elsif (left_shift(7) or left_shift(6) or left_shift(5)) = '1' then
          shift_fraction := "0" & left_fraction;
        else
          shift_fraction := "XXXXXXXXXXXXXXXXXXXXXXXXX";
        end if;
      else
        shift_sign := right_sign;
        shift_exponent := right_exponent;
        if (right_shift(7) or right_shift(6) or right_shift(5)) = '0' then
          shift_fraction := "0" & std_logic_vector(shr(unsigned(left_fraction), unsigned(right_shift(4 downto 0))));
          if (left_sign xor right_sign) = '0' then
            shift_fraction := ("0" & right_fraction) + shift_fraction;
          elsif (left_sign xor right_sign) = '1' then      
            shift_fraction := ("0" & right_fraction) - shift_fraction;
          else
            shift_fraction := "XXXXXXXXXXXXXXXXXXXXXXXXX";
          end if;
        elsif (right_shift(7) or right_shift(6) or right_shift(5)) = '1' then
          shift_fraction := "0" & right_fraction;
        else
          shift_fraction := "XXXXXXXXXXXXXXXXXXXXXXXXX";
        end if;
      end if;
      case funct is
        when "1000" =>
          sign_buf <= shift_sign;
          exponent_buf <= "0" & shift_exponent;
          fraction_buf <= shift_fraction & "00000000000000000000000";
        when "1001" =>
          sign_buf <= shift_sign;
          exponent_buf <= "0" & shift_exponent;
          fraction_buf <= shift_fraction & "00000000000000000000000";
        when "1010" =>
          sign_buf <= left_sign xor right_sign;
          exponent_buf <= ("0" & left_exponent) + ("0" & right_exponent) - 128;
          fraction_buf <= unsigned(left_fraction) * unsigned(right_fraction);
        when others =>
          sign_buf <= 'X';
          exponent_buf <= "XXXXXXXXX";
          fraction_buf <= "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX";
      end case;
    end if;
  end process;
  
 process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      if complete_buf = '1' then
        count <= "000";
      elsif (count /= "100" or funct /= "1110" or USBRXFX = '0') and (count /= "010" or funct /= "1111" or USBTXEX = '0') then
        count <= count + 1;
      end if;      
    end if;
  end process;


  process (CLK)
  begin  -- process
    if CLK'event and CLK = '1' then  -- rising clock edge
      case funct is
        when "0000" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "0001" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "0010" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "0011" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "0100" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "0101" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "0110" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "0111" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "1000" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "1001" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "1010" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "1011" =>
          USBRDX <= '1';
          USBWR <= '0';
        when "1110" =>
          case count is
            when "000" => USBRDX <= '1';
            when "001" => USBRDX <= '1';
            when "010" => USBRDX <= '1';
            when "011" => USBRDX <= '1';
            when "100" => USBRDX <= '1';
            when "101" => USBRDX <= '0';
            when "110" => USBRDX <= '0';
            when "111" => USBRDX <= '0';
            when others => USBRDX <= 'X';        
          end case;
          USBWR <= '0';
        when "1111" =>
          USBRDX <= '1';
          case count is
            when "000" => USBWR <= '0';
            when "001" => USBWR <= '0';
            when "010" => USBWR <= '0';
            when "011" => USBWR <= '1';
            when "100" => USBWR <= '1';
            when "101" => USBWR <= '1';
            when "110" => USBWR <= '0';
            when "111" => USBWR <= '0';
            when others => USBWR <= 'X';        
          end case;
        when others =>
          USBRDX <= 'X';
          USBWR <= 'X';
      end case;
    end if;
  end process;
  
end Behavioral;
